/*
	nlClassEyes.m

	This class is the controller class of the Eyes NiteLite module.  Use this code as an 
	example of how to create a NiteLite module.

	You may freely copy, distribute, and reuse the code in this example.
	Metrosoft disclaims any warranty of any kind, expressed or  implied, as to its
	fitness for any particular use.
*/

#import "nlClassEyes.h"
#import <MetroTools/NiteLite/NiteLiteSupport.h>


/* module version */
#define EYESMODVERSION	100

@implementation nlClassEyes

/* interface methods */
- setEyesCount:sender
{
	nEyes = [(Control*)sender intValue];
	[self sendChanged];		/* notify NiteLite that we have changed our settings */
	
    return self;
}

- setSounds:sender
{
	doSounds = [sender state] > 0;
	[self sendChanged];		/* notify NiteLite that we have changed our settings */
	
    return self;
}

/* NiteLite nlClassModule Methods */

/* return YES here to notify NiteLite that this module uses sounds.  This way, NiteLite will 
	display the volume controls for this module.  If you return NO here, NiteLite will not 
	display the volume controls.
*/
- (BOOL)usesSound
{
	return( YES );
}

/* here is where we read in our saved settings.  You should use the passed version number to
	handle previous versions of the saved settings data.  Don't forget to set the UI objects to
	the passed settings.
*/ 
- loadSettings:(NXTypedStream*)tsp:(int)version
{
	if( version == EYESMODVERSION )
	{
		NXReadType( tsp, "i", &nEyes );
		NXReadType( tsp, "c", &doSounds );
		
		/* set user interface to settings */
		[(Control*)countSlider setIntValue:nEyes];
		[soundsCheck setState:(doSounds?1:0)];
	}
	else
		[self defaultSettings];
		
	return( self );
}

/* this is where we write out our user settings.  NiteLite takes care of saving this out to
	the proper preferences file (depending on whether it is for foreground or background).
*/
- writeSettings:(NXTypedStream*)tsp
{
	NXWriteType( tsp, "i", &nEyes );
	NXWriteType( tsp, "c", &doSounds );

	return( self );
}

/* set our user settings here to their default values.  Don't forget to reset the UI objects. */
- defaultSettings
{
	nEyes = 5;
	doSounds = NO;
	
	/* set user interface to settings */
	[(Control*)countSlider setIntValue:nEyes];
	[soundsCheck setState:(doSounds?1:0)];
	
	return( self );
}

/* pass back our current module version.  This is saved off with the user settings by NiteLite
	and is passed back into us for verification of data version in loadSettings::.
*/
- (int)version
{
	return( EYESMODVERSION );
}

/* This method is called when drawing is about to begin.  It is a good idea to wait at least
	until here to load your display data (as opposed to doing it in initSelf:).  The parameter
	'drawData' is controlled by NiteLite.  The reason we are storing our drawing parameters in 
	'drawData->userData' is because there may be multiple places that this module may be drawing
	(because of MultiModules), so you can't store the display control variables as instance 
	variables, since they may be used multiple times.
*/
- prepDraw:(drawParamRec*)drawData:(BOOL)inFore
{
	int		i;
	
	/* allocate userData for our display control variables */
	dispData = (EyesDispRec*)malloc( sizeof(EyesDispRec) );
	drawData->userData = dispData;
	
	/* initialize the sound and image lists */
	for( i=0; i<NSOUNDS; i++ )
		soundList[i] = NULL;
	
	for( i=0; i<NEYES; i++ )
		eyes_a[i] = eyes_b[i] = eyes_c[i] = NULL;
	
	/* initialize our display control variables */
	dispData->curNEyes = 0;
	dispData->lastSoundTime = 0;
	dispData->lastEyeTime = 0;
	
	return( self );
}

/* This method is called for each animated frame.  Try to do as little drawing as possible here
	at one time, instead, spreading it out between multiple frames.  This way we don't use TOO 
	much processing time, and also the user can abort the display in a more timely fashion.
*/
- drawFrame:(drawParamRec*)drawData:(BOOL*)didDraw:(NXColor)backColor
{
	unsigned	curTime;
	int			i, j, index, nTries;
	EyeRec		*eye;
	NXRect		r, r1, bounds;
	id			img;
	BOOL		overlap;
	
	/* Only return YES for didDraw if we actually draw.  This way, NiteLite doesn't have to do
		unnecessary FlushWindow calls.
	*/
	*didDraw = NO;
	
	/* Save off our display control parameters so that we can access them in other methods we
		may call.
	*/
	dispData = drawData->userData;
	
	/* get timing value */
	curTime = currentTimeInMs();
	
	/* get the bounds we are drawing in */
	[drawData->theView getBounds:&bounds];
	
	/* change existing eyes */
	for( i=0; i<dispData->curNEyes; i++ )
	{
		eye = &(dispData->eyes[i]);
		
		/* check to see if it's time to change eyes... */
		if( !eye->inBlink && ((eye->lastBlink+EYEDELAY) < curTime || (curTime < eye->lastBlink)) )
		{
			if( !(random() % 5) )		/* change every 1 - 5 secs */
			{
				if( (eye->blinkCount >= MAXBLINKS) || 
					((eye->blinkCount > MINBLINKS) && !(random() % 3)) )
				{
					/* let's change our eyes and move them */
					if( eyes_a[eye->eyeIndex] )
						[eyes_a[eye->eyeIndex] getSize:&r.size];
					else
						r.size.width = r.size.height = 1;
					
					/* erase current eye */
					r.origin = eye->eyePos;
					NXSetColor( backColor );
					NXRectFill( &r );
					
					/* get new eye */
					eye->eyeIndex = random() % NEYES;
					[self loadEye:eye->eyeIndex];
					
					if( eyes_a[eye->eyeIndex] )
						[eyes_a[eye->eyeIndex] getSize:&r.size];
					else
						r.size.width = r.size.height = 1;
						
					eye->lastBlink = 0;
					eye->blinkCount = 1;
					eye->blinkIndex = 3;
					eye->inBlink = YES;
					
					/* try to position new eye so it is NOT overlapping another eye */
					nTries = -1;
					do
					{
						nTries++;
						overlap = NO;
						r.origin.x = random() % (int)(bounds.size.width - r.size.width);
						r.origin.y = random() % (int)(bounds.size.height - r.size.height);
						for( j=0; !overlap && j<dispData->curNEyes; j++ )
						{
							if( j != i )
							{
								r1.origin = dispData->eyes[j].eyePos;
								if( eyes_a[dispData->eyes[j].eyeIndex] )
									[eyes_a[dispData->eyes[j].eyeIndex] getSize:&r1.size];
								else
									r1.size.width = r1.size.height = 1;
								
								if( NXIntersectsRect( &r, &r1 ) )
									overlap = YES;
							}
						}
					} while( overlap && nTries < 100 );
					
					eye->eyePos = r.origin;
				}
				else
				{
					/* just start a blink */
					eye->inBlink = YES;
					eye->blinkCount++;
					eye->blinkIndex = 0;
				}
			}
			else
				eye->lastBlink = curTime;
		}
	}
	
	/* set up new eyes */
	if( dispData->curNEyes < MAXEYES && dispData->curNEyes < nEyes && 
		((dispData->lastEyeTime+EYEDELAY) < curTime || (curTime < dispData->lastEyeTime)) )
	{
		if( !(random() % 5) || !dispData->curNEyes )		/* change every 1 - 5 secs */
		{
			/* load in new eye */
			eye = &(dispData->eyes[dispData->curNEyes]);
			dispData->curNEyes++;
			eye->eyeIndex = random() % NEYES;
			[self loadEye:eye->eyeIndex];
			
			if( eyes_a[eye->eyeIndex] )
				[eyes_a[eye->eyeIndex] getSize:&r.size];
			else
				r.size.width = r.size.height = 0;
				
			eye->lastBlink = 0;
			eye->blinkCount = 1;
			eye->blinkIndex = 3;
			eye->inBlink = YES;
			
			/* try to position new eye so it is NOT overlapping another eye */
			nTries = -1;
			do
			{
				nTries++;
				overlap = NO;
				r.origin.x = random() % (int)(bounds.size.width - r.size.width);
				r.origin.y = random() % (int)(bounds.size.height - r.size.height);
				for( i=0; !overlap && i<dispData->curNEyes; i++ )
				{
					if( i != eye->eyeIndex )
					{
						r1.origin = dispData->eyes[i].eyePos;
						if( eyes_a[dispData->eyes[i].eyeIndex] )
							[eyes_a[dispData->eyes[i].eyeIndex] getSize:&r1.size];
						else
							r1.size.width = r1.size.height = 1;
						
						if( NXIntersectsRect( &r, &r1 ) )
							overlap = YES;
					}
				}
			} while( overlap && nTries < 100 );
			
			eye->eyePos = r.origin;
		}
		
		dispData->lastEyeTime = curTime;
	}
	
	/* draw eyes */
	for( i=0; i<dispData->curNEyes; i++ )
	{
		eye = &(dispData->eyes[i]);
		if( eye->inBlink && ((eye->lastBlink+EYEANIMTIME)<curTime || (curTime<eye->lastBlink)) )
		{
			/* erase eyes */
			[eyes_a[eye->eyeIndex] getSize:&r.size];
			r.origin = eye->eyePos;
			NXSetColor( backColor );
			NXRectFill( &r );
			
			/* calc which animation frame to display for eye */
			switch( eye->blinkIndex++ )
			{
				case 0:
				case 6:
					img = eyes_a[eye->eyeIndex];
					break;
				case 1:
				case 5:
					img = eyes_b[eye->eyeIndex];
					break;
				case 2:
				case 4:
					img = eyes_c[eye->eyeIndex];
					break;
				default:
					img = NULL;
					break;
			}
			
			/* are we done with the blink? */
			if( eye->blinkIndex > 6 )
				eye->inBlink = NO;
			
			/* display eye */
			if( img )
				[img composite:NX_SOVER toPoint:(const NXPoint*)&r.origin];
			
			/* notify NiteLite that we drew */
			*didDraw = YES;
			eye->lastBlink = curTime;
		}
	}
	
	/* play sound */
	if( doSounds && 
		((dispData->lastSoundTime + SOUNDDELAY) < curTime || (curTime < dispData->lastSoundTime)) )
	{
		/* Note that when we call playSound, we pass different IDs for the sounds depending on
			if they are the main sound or not.  This way they can be played synchronously with
			each other, but we won't end up with multiple intances of the same sound queued up.
		*/
		if( NSOUNDS > 1 && !(random() % 10) )		/* non-cricket sounds every 2 - 20 seconds */
		{
			index = (random() % (NSOUNDS-1)) + 1;
			[self loadSound:index];
			if( soundList[index] )
				[self playSound:soundList[index]:2000];
		}
		else if( !(random() % 3) )					/* cricket sounds every 2 - 6 secs */
		{
			[self loadSound:0];
			if( soundList[0] )
				[self playSound:soundList[0]:1000];
		}
		
		dispData->lastSoundTime = curTime;
	}
	
	return( self );
}

/* This method is called after drawing is finished.  Any things that were loaded at prepDraw::
	or during drawing should be disposed of here.  Don't forget to get rid of your userData!
*/ 
- endDraw:(drawParamRec*)drawData
{
	int		i;
	
	/* dispose of your 'userData' display control data */
	if( drawData->userData )
		free( drawData->userData );
	
	/* stop any currently playing (or queued up) sounds */
	[self stopSound:1000];
	[self stopSound:2000];
	
	/* get rid of loaded sounds */
	for( i=0; i<NSOUNDS; i++ )
	{
		if( soundList[i] )
		{
			[soundList[i] free];
			soundList[i] = NULL;
		}
	}
	
	/* get rid of loaded eye images */
	for( i=0; i<NEYES; i++ )
	{
		if( eyes_a[i] )
			[eyes_a[i] free];
		if( eyes_b[i] )
			[eyes_b[i] free];
		if( eyes_c[i] )
			[eyes_c[i] free];
			
		eyes_a[i] = NULL;
		eyes_b[i] = NULL;
		eyes_c[i] = NULL;
	}
	
	return( self );
}

/* support methods */

/* this method uses retrieveSound: to load in a sound */
- loadSound:(int)index
{
	char	name[20];
	
	if( !soundList[index] )
	{
		sprintf( name, "animal%d", index+1 );
		soundList[index] = [self retrieveSound:name];
	}
	
	return( self );
}

/* this method loads in an eye image sequence from the module's NXBundle */
- loadEye:(int)index
{
	char	imgName[20], path[MAXPATHLEN+1];
	
	/* load eye frame 1 */
	if( !eyes_a[index] )
	{
		sprintf( imgName, "eyes%da", index+1 );
		[modBundle getPath:path forResource:imgName ofType:"tiff"];
		eyes_a[index] = [[NXImage alloc] initFromFile:path];
	}
	
	/* load eye frame 2 */
	if( !eyes_b[index] )
	{
		sprintf( imgName, "eyes%db", index+1 );
		[modBundle getPath:path forResource:imgName ofType:"tiff"];
		eyes_b[index] = [[NXImage alloc] initFromFile:path];
	}
	
	/* load eye frame 3 */
	if( !eyes_c[index] )
	{
		sprintf( imgName, "eyes%dc", index+1 );
		[modBundle getPath:path forResource:imgName ofType:"tiff"];
		eyes_c[index] = [[NXImage alloc] initFromFile:path];
	}
	
	return( self );
}



@end
